@@ -5,10 +5,10 @@
  *		     init [0123456SsQqAaBbCc]
  *		  telinit [0123456SsQqAaBbCc]
  *
- * Version:	@(#)init.c  2.71  05-May-1997  MvS
+ * Version:	@(#)init.c  2.73  03-Jan-1998  MvS
  *
  *		This file is part of the sysvinit suite,
- *		Copyright 1991-1997 Miquel van Smoorenburg.
+ *		Copyright 1991-1998 Miquel van Smoorenburg.
  *
  *		This program is free software; you can redistribute it and/or
  *		modify it under the terms of the GNU General Public License
@@ -52,9 +52,9 @@
 				sigaction(sig, &sa, NULL);
 
 /* Version information */
-char *Version = "@(#) init 2.71 05-May-1997 MvS";
-char *bootmsg = "version 2.71 booting";
-#define E_VERSION "INIT_VERSION=sysvinit-2.71"
+char *Version = "@(#) init 2.73 03-Jan-1998 MvS";
+char *bootmsg = "version 2.73 booting";
+#define E_VERSION "INIT_VERSION=sysvinit-2.73"
 
 CHILD *family = NULL;		/* The linked list of all entries */
 CHILD *newFamily = NULL;	/* The list after inittab re-read */
@@ -83,6 +83,7 @@ int maxproclen;			/* Maximal length of argv[0] */
 struct utmp utproto;		/* Only used for sizeof(utproto.ut_id) */
 char *console_dev = CONSOLE;	/* Console device. */
 int wrote_iosave = 0;		/* Did we write ioctl.save yet? */
+int pipe_fd = -1;		/* /dev/initctl */
 
 /* Macro to see if this is a special action */
 #define ISPOWER(i) ((i) == POWERWAIT || (i) == POWERFAIL || \
@@ -282,7 +283,8 @@ void set_term(int how)
   (void) ioctl(fd, TCGETS, &tty);
   if ((fd2 = open(IOSAVE, O_WRONLY|O_CREAT|O_TRUNC, 0600)) < 0) {
 	if (errno != EROFS)
-		log(L_VB, "can't open(%s, O_WRONLY): %s", IOSAVE, sys_errlist[errno]);
+		log(L_VB, "can't open(%s, O_WRONLY): %s",
+			IOSAVE, strerror(errno));
 	(void) close(fd);
 	(void) close(fd2);
 	return;
@@ -384,7 +386,7 @@ int any(char *s1, char *s2)
  *	This function is too long and indents too deep.
  *
  */
-int spawn(CHILD *ch)
+int spawn(CHILD *ch, int *res)
 {
   char *args[16];		/* Argv array */
   char buf[136];		/* Line buffer */
@@ -397,6 +399,9 @@ int spawn(CHILD *ch)
   char i_prev[] = "PREVLEVEL=x";/* Previous runlevel. */
   char i_cons[32];		/* console device. */
   pid_t pid, pgrp;		/* child, console process group. */
+  sigset_t nmask, omask;	/* For blocking SIGCHLD */
+
+  *res = -1;
 
   /* Skip '+' if it's there */
   if (proc[0] == '+') proc++;
@@ -486,11 +491,22 @@ int spawn(CHILD *ch)
   }
   args[0] = args[1];
   while(1) {
+	/*
+	 *	Block sigchild while forking.
+	 */
+	sigemptyset(&nmask);
+	sigaddset(&nmask, SIGCHLD);
+	sigprocmask(SIG_BLOCK, &nmask, &omask);
+
 	if ((pid = fork()) == 0) {
+
 		close(0);
 		close(1);
 		close(2);
-		close(PIPE_FD);
+		if (pipe_fd >= 0) close(pipe_fd);
+
+  		sigprocmask(SIG_SETMASK, &omask, NULL);
+
 		/* Now set RUNLEVEL and PREVLEVEL */
 		sprintf(i_cons, "CONSOLE=%s", console_dev);
 		i_lvl[9]   = thislevel;
@@ -558,7 +574,7 @@ int spawn(CHILD *ch)
   		} else {
 			if ((f = serial_open(console_dev, O_RDWR|O_NOCTTY)) < 0) {
 				log(L_VB, "open(%s): %s", console_dev,
-					sys_errlist[errno]);
+					strerror(errno));
 				f = open("/dev/null", O_RDWR);
 			}
 			dup(f);
@@ -568,8 +584,10 @@ int spawn(CHILD *ch)
   		/* Reset all the signals */
   		for(f = 1; f < _NSIG; f++) signal(f, SIG_DFL);
   		execvp(args[1], args + 1);
-		/* Is this a bug in execvp? It does _not_ execute shell
-		 * scripts (/etc/rc !), so we try again with 'sh -c exec ...'
+		/*
+		 *	Is this a bug in execvp? It does _not_ execute shell
+		 *	scripts (/etc/rc !), so we try again with
+		 *	'sh -c exec ...'
 		 */
 		if (errno == ENOEXEC) {
   			args[1] = SHELL;
@@ -583,6 +601,9 @@ int spawn(CHILD *ch)
   		log(L_VB, "cannot execute \"%s\"", args[1]);
   		exit(1);
   	}
+	*res = pid;
+  	sigprocmask(SIG_SETMASK, &omask, NULL);
+
 #if DEBUG
 	log(L_VB, "Started id %s (pid %d)", ch->id, pid);
 #endif
@@ -620,7 +641,7 @@ void startup(CHILD *ch)
 	case ONDEMAND:
 	case RESPAWN:
   		ch->flags |= RUNNING;
-  		if ((ch->pid = spawn(ch)) < 0) break;
+  		if (spawn(ch, &(ch->pid)) < 0) break;
 		/* Do NOT log if process field starts with '+' */
   		if (ch->process[0] != '+')
 			write_utmp_wtmp("", ch->id, ch->pid, INIT_PROCESS, "");
@@ -643,7 +664,9 @@ char *get_part(char *str, int tok)
   return(q);
 }
 
-/* Read the inittab file. */
+/*
+ *	Read the inittab file.
+ */
 void read_inittab(void)
 {
   FILE *fp;			/* The INITTAB file */
@@ -651,6 +674,7 @@ void read_inittab(void)
   char err[64];			/* Error message. */
   char *id, *rlevel,
        *action, *process;	/* Fields of a line */
+  char *p;
   CHILD *ch, *old, *i;		/* Pointers to CHILD structure */
   CHILD *head = NULL;		/* Head of linked list */
   int lineNo = 0;		/* Line number in INITTAB file */
@@ -673,16 +697,21 @@ void read_inittab(void)
   log(L_VB, "Reading inittab");
 #endif
 
-  /* Open INITTAB */
+  /*
+   *	Open INITTAB and real line by line.
+   */
   if ((fp = fopen(INITTAB, "r")) == NULL)
 	log(L_VB, "No inittab file found");
 
-  /* Read INITTAB line by line */
   while(!done) {
-	/* Add single user shell entry as last one. */
+	/*
+	 *	Add single user shell entry at the end.
+	 */
 	if (fp == NULL || fgets(buf, 255, fp) == NULL) {
 		done = 1;
-		/* See if we have a single user entry. */
+		/*
+		 *	See if we have a single user entry.
+		 */
 		for(old = newFamily; old; old = old->next)
 			if (strcmp(old->rlevel, "S") == 0) break;
 		if (old == NULL)
@@ -691,17 +720,24 @@ void read_inittab(void)
 			continue;
 	}
 	lineNo++;
-	/* Skip comments and empty lines */
-	if (buf[0] == '#' || buf[0] == '\n') continue;
+	/*
+	 *	Skip comments and empty lines
+	 */
+	for(p = buf; *p == ' ' || *p == '\t'; p++)
+		;
+	if (*p == '#' || *p == '\n') continue;
 
-	/* Decode the fields */
-	id =      get_part(buf,  ':');
+	/*
+	 *	Decode the fields
+	 */
+	id =      get_part(p,  ':');
 	rlevel =  get_part(NULL, ':');
 	action =  get_part(NULL, ':');
 	process = get_part(NULL, '\n');
 
-	/* Check if syntax is OK. Be very verbose here, to
-	 * avoid newbie postings on comp.os.linux.setup :)
+	/*
+	 *	Check if syntax is OK. Be very verbose here, to
+	 *	avoid newbie postings on comp.os.linux.setup :)
 	 */
 	err[0] = 0;
 	if (!id || !*id) strcpy(err, "missing id field");
@@ -723,7 +759,9 @@ void read_inittab(void)
 		continue;
 	}
   
-	/* Decode the "action" field */
+	/*
+	 *	Decode the "action" field
+	 */
 	actionNo = -1;
 	for(f = 0; actions[f].name; f++)
 		if (strcasecmp(action, actions[f].name) == 0) {
@@ -736,7 +774,9 @@ void read_inittab(void)
 		continue;
 	}
 
-	/* See if the id field is unique */
+	/*
+	 *	See if the id field is unique
+	 */
 	for(old = newFamily; old; old = old->next) {
 		if(strcmp(old->id, id) == 0 && strcmp(id, "~~")) {
 			log(L_VB, "%s[%d]: duplicate ID field \"%s\"",
@@ -746,14 +786,18 @@ void read_inittab(void)
 	}
 	if (old) continue;
 
-	/* Allocate a CHILD structure */
+	/*
+	 *	Allocate a CHILD structure
+	 */
 	if ((ch = malloc(sizeof(CHILD))) == NULL) {
 		log(L_VB, "out of memory");
 		continue;
 	}
 	memset(ch, 0, sizeof(CHILD));
 
-	/* And fill it in. */
+	/*
+	 *	And fill it in.
+	 */
 	ch->action = actionNo;
 	strncpy(ch->id, id, sizeof(utproto.ut_id) + 1); /* Hack for different libs. */
 	strncpy(ch->process, process, 127);
@@ -769,26 +813,34 @@ void read_inittab(void)
 			strcpy(ch->rlevel, "S0123456789");
 	}
 	/*
-	 * We have the fake runlevel '#' for SYSINIT  and
-	 * '*' for BOOT and BOOTWAIT.
+	 *	We have the fake runlevel '#' for SYSINIT  and
+	 *	'*' for BOOT and BOOTWAIT.
 	 */
 	if (ch->action == SYSINIT) strcpy(ch->rlevel, "#");
 	if (ch->action == BOOT || ch->action == BOOTWAIT)
 		strcpy(ch->rlevel, "*");
 
-	/* Now add it to the linked list. Special for powerfail. */
+	/*
+	 *	Now add it to the linked list. Special for powerfail.
+	 */
 	if (ISPOWER(ch->action)) {
 
-		/* Disable by default */
+		/*
+		 *	Disable by default
+		 */
 		ch->flags |= XECUTED;
 
-		/* Tricky: insert at the front of the list.. */
+		/*
+		 *	Tricky: insert at the front of the list..
+		 */
 		old = NULL;
 		for(i = newFamily; i; i = i->next) {
 			if (!ISPOWER(i->action)) break;
 			old = i;
 		}
-		/* Now add after entry "old" */
+		/*
+		 *	Now add after entry "old"
+		 */
 		if (old) {
 			ch->next = i;
 			old->next = ch;
@@ -799,7 +851,9 @@ void read_inittab(void)
 			if (ch->next == NULL) head = ch;
 		}
 	} else {
-		/* Just add at end of the list */
+		/*
+		 *	Just add at end of the list
+		 */
 		if (ch->action == KBREQUEST) ch->flags |= XECUTED;
 		ch->next = NULL;
 		if (head)
@@ -809,19 +863,23 @@ void read_inittab(void)
 		head = ch;
 	}
 
-	/* Walk through the old list comparing id fields */
+	/*
+	 *	Walk through the old list comparing id fields
+	 */
 	for(old = family; old; old = old->next)
 		if (strcmp(old->id, ch->id) == 0) {
 			old->new = ch;
 			break;
 		}
   }
-  /* We're done. */
+  /*
+   *	We're done.
+   */
   if (fp) fclose(fp);
 
   /*
-   * Loop through the list of children, and see if they need to
-   * be killed. 
+   *	Loop through the list of children, and see if they need to
+   *	be killed. 
    */
 
 #if DEBUG
@@ -832,24 +890,36 @@ void read_inittab(void)
     for(ch = family; ch; ch = ch->next) {
 	ch->flags &= ~KILLME;
 
-	/* Is this line deleted? */
+	/*
+	 *	Is this line deleted?
+	 */
 	if (ch->new == NULL) ch->flags |= KILLME;
 
-#if 0 /* XXX - This is more compatible with real sysv. */
-	/* See if the entry changed. Yes: kill anyway */
-	if (ch->new && (strcmp(ch->process, ch->new->process) ||
-		ch->action != ch->new->action)) ch->flags |= KILLME;
-#endif
+	/*
+	 *	If the entry has changed, kill it anyway. Note that
+	 *	we do not check ch->process, only the "action" field.
+	 *	This way, you can turn an entry "off" immidiately, but
+	 *	changes in the command line will only become effective
+	 *	after the running version has exited.
+	 */
+	if (ch->new && ch->action != ch->new->action) ch->flags |= KILLME;
 
-	/* Only BOOT processes may live in all levels */
+	/*
+	 *	Only BOOT processes may live in all levels
+	 */
 	if (ch->action != BOOT &&
 	    strchr(ch->rlevel, runlevel) == NULL) {
-		/* Ondemand procedures live always except in single user */
+		/*
+		 *	Ondemand procedures live always,
+		 *	except in single user
+		 */
 		if (runlevel == 'S' || !(ch->flags & DEMAND))
 			ch->flags |= KILLME;
 	}
 
-	/* Now, if this process may live note so in the new list */
+	/*
+	 *	Now, if this process may live note so in the new list
+	 */
 	if ((ch->flags & KILLME) == 0) {
 		ch->new->flags  = ch->flags;
 		ch->new->pid    = ch->pid;
@@ -858,7 +928,9 @@ void read_inittab(void)
 	}
 
 
-	/* Is this process still around? */
+	/*
+	 *	Is this process still around?
+	 */
 	if ((ch->flags & RUNNING) == 0) {
 		ch->flags &= ~KILLME;
 		continue;
@@ -881,11 +953,14 @@ void read_inittab(void)
 	}
 	talk = 0;
 	
-	/* And process the next member of our family */
     }
-    /* See if we have to wait 5 seconds */
+    /*
+     *	See if we have to wait 5 seconds
+     */
     if (foundOne && round == 0) {
-	/* Yup, but check every second if we still have children. */
+	/*
+	 *	Yup, but check every second if we still have children.
+	 */
 	for(f = 0; f < sltime; f++) {
 		for(ch = family; ch; ch = ch->next) {
 			if (!(ch->flags & KILLME)) continue;
@@ -893,7 +968,9 @@ void read_inittab(void)
 				break;
 		}
 		if (ch == NULL) {
-			/* No running children, skip SIGKILL */
+			/*
+			 *	No running children, skip SIGKILL
+			 */
 			round = 1;
 			foundOne = 0; /* Skip the sleep below. */
 			break;
@@ -922,7 +999,9 @@ void read_inittab(void)
 		}
 	}
 
-  /* Both rounds done; clean up the list. */
+  /*
+   *	Both rounds done; clean up the list.
+   */
   sigemptyset(&nmask);
   sigaddset(&nmask, SIGCHLD);
   sigprocmask(SIG_BLOCK, &nmask, &omask);
@@ -936,22 +1015,34 @@ void read_inittab(void)
   sigprocmask(SIG_SETMASK, &omask, NULL);
 
 #ifdef INITLVL
-  /* Dispose of INITLVL file. */
+  /*
+   *	Dispose of INITLVL file.
+   */
   if (lstat(INITLVL, &st) >= 0 && S_ISLNK(st.st_mode)) {
-	/* INITLVL is a symbolic link, so just truncate the file. */
+	/*
+	 *	INITLVL is a symbolic link, so just truncate the file.
+	 */
 	close(open(INITLVL, O_WRONLY|O_TRUNC));
   } else {
-	/* Delete INITLVL file. */
+	/*
+	 *	Delete INITLVL file.
+	 */
   	unlink(INITLVL);
   }
 #endif
 #ifdef INITLVL2
-  /* Dispose of INITLVL2 file. */
+  /*
+   *	Dispose of INITLVL2 file.
+   */
   if (lstat(INITLVL2, &st) >= 0 && S_ISLNK(st.st_mode)) {
-	/* INITLVL2 is a symbolic link, so just truncate the file. */
+	/*
+	 *	INITLVL2 is a symbolic link, so just truncate the file.
+	 */
 	close(open(INITLVL2, O_WRONLY|O_TRUNC));
   } else {
-	/* Delete INITLVL2 file. */
+	/*
+	 *	Delete INITLVL2 file.
+	 */
   	unlink(INITLVL2);
   }
 #endif
@@ -1017,9 +1108,6 @@ int ask_runlevel()
 
 
   set_term(0);
-  /*
-   *	FIXME: should we open the tty O_NONBLOCK and CLOCAL mode?
-   */
   fd = serial_open(console_dev, O_RDWR|O_NOCTTY);
 
   if (fd < 0) return('S');
@@ -1287,7 +1375,6 @@ void fifo_new_level(int level)
  */
 void check_init_fifo(void)
 {
-  static int pipe_fd = -1;
   struct init_request request;
   int n;
   fd_set fds;
@@ -1382,6 +1469,7 @@ void boot_transitions()
   static int newlevel = 0;
   static int loglevel;
   static int oldlevel;
+  static int warn = 1;
 
   /* Check if there is something to wait for! */
   for( ch = family; ch; ch = ch->next )
@@ -1423,6 +1511,7 @@ void boot_transitions()
 				loglevel = newlevel;
 			runlevel = newlevel;
 			did_boot = 1;
+			warn = 1;
 			break;
 		case 'S': /* Ended SU mode */
 		case 's':
@@ -1441,13 +1530,16 @@ void boot_transitions()
 				runlevel = newlevel;
 				oldlevel = 'S';
 			}
+			warn = 1;
 			for(ch = family; ch; ch = ch->next)
 			    if (strcmp(ch->rlevel, "S") == 0)
 				ch->flags &= ~(FAILING|WAITING|XECUTED);
 			break;
 		default:
+			if (warn)
+			  log(L_VB, "no more processes left in this runlevel");
+			warn = 0;
 			loglevel = -1;
-			log(L_VB, "no more processes left in this runlevel");
 			if (got_signals == 0)
 				check_init_fifo();
 			break;
@@ -1565,6 +1657,16 @@ void process_signals()
 		DELSET(got_signals, SIGHUP);
 	}
   }
+  if (ISMEMBER(got_signals, SIGUSR1)) {
+	/*
+	 *	SIGUSR1 means close and reopen /dev/initctl
+	 */
+#if DEBUG
+	log(L_VB, "got SIGHUP");
+#endif
+	close(pipe_fd);
+	pipe_fd = -1;
+  }
 }
 
 /*
@@ -1589,25 +1691,6 @@ int init_main()
   }
 #endif
 
-#ifdef ROOTFS
-  /* Close all file descriptors. */
-  close(0);
-  close(1);
-  close(2);
-
-  /* Mount new root file system on /root */
-  if (mount(ROOTFS, "/root", "ext2", 0, 0) < 0) {
-	log(L_VB, "mount(%s, /root): %s", ROOTFS, sys_errlist[errno]);
-	while(1) pause();
-  }
-
-  /* Now do the chroot (needs kernel patch). */
-  if (chroot("/root") < 0) {
-	log(L_VB, "chroot(/root): %s", sys_errlist[errno]);
-	while(1) pause();
-  }
-#endif
-
   /* Tell the kernel to send us SIGINT when CTRL-ALT-DEL is pressed */
   init_reboot(BMAGIC_SOFT);
 
@@ -1629,11 +1712,24 @@ int init_main()
   SETSIG(sa, SIGINT,   signal_handler);
   SETSIG(sa, SIGPWR,   signal_handler);
   SETSIG(sa, SIGWINCH, signal_handler);
+  SETSIG(sa, SIGUSR1,  signal_handler);
   SETSIG(sa, SIGSTOP,  stop_handler);
   SETSIG(sa, SIGTSTP,  stop_handler);
   SETSIG(sa, SIGCONT,  cont_handler);
   SETSIG(sa, SIGSEGV,  segv_handler);
 
+  /*
+   *	NOTE: instead of calling siginterrupt, we could just as
+   *	well install the signal handler without SA_RESTART in the
+   *	first place.
+   */
+  siginterrupt (SIGALRM,  1);
+  siginterrupt (SIGHUP,   1);
+  siginterrupt (SIGINT,   1);
+  siginterrupt (SIGPWR,   1);
+  siginterrupt (SIGUSR1,  1);
+  siginterrupt (SIGWINCH, 1);
+
   /* Close whatever files are open, and reset the console. */
   close(0);
   close(1);
@@ -1655,7 +1751,7 @@ int init_main()
   /* See if we have to start an emergency shell. */
   if (emerg_shell) {
 	SETSIG(sa, SIGCHLD, SIG_DFL);
-	if ((f = spawn(&ch_emerg)) > 0) {
+	if (spawn(&ch_emerg, &f) > 0) {
 		while(wait(&st) != f)
 			;
 	}
@@ -1804,6 +1900,7 @@ int main(int argc, char *argv[])
 
   /* Make sure we don't hang on opening /etc/init.fifo */
   signal(SIGALRM, signal_handler);
+  siginterrupt (SIGALRM, 1);
   alarm(3);
   if ((fd = open(INIT_FIFO, O_WRONLY)) >= 0 &&
 	    write(fd, &request, sizeof(request)) == sizeof(request)) {
